// Node modules
import AbstractClientStore from 'express-brute/lib/AbstractClientStore';
import moment from 'moment';
import Sequelize from 'sequelize';

/**
 * A store implementing express-brute AbstractClientStore functionality.
 *
 * Allows for a Sequelize connection to a MySQL database in order to store
 * results of brute force attacks.
 */
class Store {
  /**
   * The constructor for the store.
   *
   * @param  {Object}    sequelize A instance of a Sequelize object.
   * @param  {String}    table     The name of the table to store the results in.
   * @param  {Options}   options   Additional options to pass to the store.
   *
   * @return {Object}              The store object.
   */
  constructor(sequelize, table, options) {
    AbstractClientStore.apply(this);

    this.defaults = {
      prefix: '',
      logging: false,
    };

    this.options = Object.assign({}, this.defaults, options);

    // Define the Sequelize model
    this.table = sequelize.define(table, {
      _id: {
        type: Sequelize.STRING,
        unique: true,
      },
      expires: {
        type: Sequelize.DATE,
      },
      firstRequest: {
        type: Sequelize.DATE,
      },
      lastRequest: {
        type: Sequelize.DATE,
      },
      count: {
        type: Sequelize.INTEGER,
      },
    });
  }

  /**
   * Migrates the database to include the table if needed.
   *
   * @return {Null} Returns null regardless of migration.
   */
  async migrate() {
    try {
      await this.table.sync();
    } catch (error) {
      // eslint-disable-next-line no-console
      console.error(`Failed to initialize Store - table ${this.table}`);
    }
  }

  /**
   * Sets a value into the table.
   *
   * @param {String} key      The key to base the _id after.
   * @param {Object} value    The value object of express-brute.
   * @param {Date}   lifetime A date in which to expire the value in.
   */
  async set(key, value, lifetime) {
    const id = this.options.prefix + key;

    // Build the expiration date
    const expiration = lifetime ? (moment().add(lifetime, 'seconds')).toDate() : null;

    const doc = await this.table.find({
      where: {
        _id: id,
      },
    });

    if (doc) {
      // eslint-disable-next-line no-underscore-dangle
      doc._id = id;
      doc.count = value.count;
      doc.lastRequest = value.lastRequest;
      doc.firstRequest = value.firstRequest;
      doc.expires = expiration;

      // Update the existing document with the new values
      return doc.save();
    }

    // Create a new document with the values
    return this.table.create({
      _id: id,
      count: value.count,
      lastRequest: value.lastRequest,
      firstRequest: value.firstRequest,
      expires: expiration,
    });
  }

  /**
   * Retrives a record from the table by the given key.
   *
   * @param  {String} key The key which to retrieve the record by.
   *
   * @return {Null} Always returns null.
   */
  async get(key) {
    const id = this.options.prefix + key;

    const doc = await this.table.find({
      where: {
        _id: id,
      },
    });

    if (doc && new Date(doc.expires).getTime() < new Date().getTime()) {
      await this.table.destroy({
        where: {
          _id: id,
        },
      });
      return null;
    }

    if (doc) {
      return {
        count: doc.count,
        lastRequest: new Date(doc.lastRequest),
        firstRequest: new Date(doc.firstRequest),
      };
    }

    return null;
  }

  /**
   * Destroys a record by the given key.
   *
   * @param  {String} key The key which to destroy the record by.
   *
   * @return {Null} Always returns null.
   */
  async reset(key) {
    const id = this.options.prefix + key;

    await this.table.destroy({
      where: {
        _id: id,
      },
    });

    return null;
  }
}

export default Store;
